static sreg_t error_noperm(ureg_t *args)
{
    return -EPERM; /* we don't do dom0 */
}

static sreg_t error_nosys(ureg_t *args)
{
    return -ENOSYS; /* not implemented */
}

static sreg_t stack_switch(ureg_t *args)
{
    xen_ctxt.kernel_ss = fix_sel(args[0]);
    xen_ctxt.kernel_sp = args[1];

#if 0 == longmode
    xen_tss.ss1  = xen_ctxt.kernel_ss;
    xen_tss.esp1 = xen_ctxt.kernel_sp;
#endif
    return 0;
}

static sreg_t update_descriptor(ureg_t *args)
{
#if longmode
    uint64_t pa = args[0];
    struct descriptor_32 desc = {
	.a = args[1] & 0xffffffff,
	.b = args[1] >> 32,
    };
#else
    uint64_t pa   = args[0] | (uint64_t)args[1] << 32;
    struct descriptor_32 desc = {
	.a = args[2],
	.b = args[3],
    };
#endif
    struct descriptor_32 *guest_gdt;
    int p, index;
    uint64_t mfn;

    mfn = addr_to_frame(pa);
    for (p = 0; p < sizeof(gdt_mfn)/sizeof(gdt_mfn[0]); p++) {
	if (mfn == gdt_mfn[p])
	    break;
    }
    if (p == sizeof(gdt_mfn)/sizeof(gdt_mfn[0])) {
	printk("%s: not in gdt: pa %" PRIx64 "\n", __FUNCTION__, pa);
	return -EINVAL;
    }
    index = addr_offset(pa) / sizeof(struct descriptor_32);

    fix_desc(&desc);
    xen_gdt[p * 512 + index] = desc;   /* update emu shadow */

    /* update guest gdt  */
    guest_gdt = map_page(pa);
    *guest_gdt = desc;
    free_page(guest_gdt);
    return 0;
}

static sreg_t fpu_taskswitch(ureg_t *args)
{
    if (args[0]) {
	write_cr0(X86_CR0_TS|read_cr0());
    } else {
	__asm__("clts");
    }
    return 0;
}

static sreg_t grant_table_op(ureg_t *args)
{
    struct gnttab_setup_table *st;
    struct gnttab_query_size  *qs;
    unsigned long *frames;
    int i, rc = 0;
    
    switch (args[0]) {
    case GNTTABOP_setup_table:
	st = (void*)args[1];
	printk("%s: setup_table %d\n", __FUNCTION__, st->nr_frames);
	if (st->nr_frames > GRANT_FRAMES_MAX) {
	    st->status = GNTST_general_error;
	} else {
	    grant_frames = st->nr_frames;
	    frames = st->frame_list.p;
	    for (i = 0; i < grant_frames; i++)
		frames[i] = EMU_MFN(grant_table) + i;
	    st->status = GNTST_okay;
	}
	break;
    case GNTTABOP_query_size:
	printk("%s: query_size\n", __FUNCTION__);
	qs = (void*)args[1];
	qs->nr_frames = grant_frames;
	qs->max_nr_frames = GRANT_FRAMES_MAX;
	qs->status = GNTST_okay;
	break;
    default:
	printk("%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	rc = -ENOSYS;
    }
    return rc;
}

static sreg_t xen_version(ureg_t *args)
{
    switch (args[0]) {
    case XENVER_version:
	return (3 << 16) | 1;

    case XENVER_get_features:
    {
	xen_feature_info_t *fi = (void*)args[1];
	fi->submap = 0;
	if (0 == fi->submap_idx)
	    fi->submap |= (1 << XENFEAT_pae_pgdir_above_4gb);
 	break;
    }

#if 0 == longmode
    case XENVER_platform_parameters:
    {
	uint32_t *ptr32 = (void*)args[1];
	*ptr32 = is_pae() ? XEN_M2P_PAE : XEN_M2P_32;
	break;
    }
#endif

    default:
	printk("%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	return -ENOSYS;
    }
    return 0;
}

static sreg_t vm_assist(ureg_t *args)
{
    int type = args[1];

    switch (args[0]) {
    case VMASST_CMD_enable:
	printk("%s: enable %d\n", __FUNCTION__, type);
	break;
    case VMASST_CMD_disable:
	printk("%s: disable %d\n", __FUNCTION__, type);
	break;
    default:
	printk("%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	return -ENOSYS;
    }
    return 0;
}

static sreg_t sched_op(ureg_t *args)
{
    switch (args[0]) {
    case SCHEDOP_yield:
	/* Hmm, some day, on SMP, we want probably do something else ... */
	halt();
	break;

    case SCHEDOP_block:
	halt();
	break;

    case SCHEDOP_shutdown:
	shutdown();
	break;

    default:
	printk("%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	return -ENOSYS;
    }
    return 0;
}

static sreg_t memory_op(ureg_t *args)
{
    int cmd = args[0] & 0x0f /* MEMOP_CMD_MASK */;
    
    switch (cmd) {
    case XENMEM_machphys_mapping:
    {
	struct xen_machphys_mapping map;
	uint32_t pg_m2p = 1024 /* pages (4 MB) */;
	void *dest = (void*)(args[1]);
#if longmode
	map.v_start = XEN_M2P_64;
#else
	if (is_pae()) {
	    map.v_start = XEN_M2P_PAE;
	} else {
	    map.v_start = XEN_M2P_32;
	}
#endif
	map.v_end   = map.v_start + frame_to_addr(pg_m2p);
	map.max_mfn = pg_m2p << (PAGE_SHIFT-3);
	if (0 != memcpy_pf(dest, &map, sizeof(map)))
	    return -EFAULT;
	return 0;
    }
    
    case XENMEM_memory_map:
	/* we have no e820 map */
	return -ENOSYS;

    default:
	printk("%s: FIXME: unknown %d\n", __FUNCTION__, cmd);
	return -ENOSYS;
    }
    return 0;
}

static sreg_t set_trap_table(ureg_t *args)
{
    struct trap_info *traps;
    struct trap_info trap;
    int i;

    if (0 == args[0]) {
	printk("%s: FIXME: clear table\n", __FUNCTION__);
	return -EINVAL;
    }

    traps = (void*)args[0];
    for (i = 0;; i++) {
	if (0 != memcpy_pf(&trap, traps+i, sizeof(trap)))
	    return -EFAULT;
	if (!trap.address)
	    break;
	trap.cs = fix_sel32(trap.cs);
	xen_ctxt.trap_ctxt[traps[i].vector] = trap;
#if 0 == longmode
	if (traps[i].vector > 0x40) {
	    /* route directly */
	    uint32_t dpl = trap.flags & 0x03;
	    xen_idt[traps[i].vector] =
		mkgate32(trap.cs, trap.address, 0x8f | (dpl << 5));
	}
#endif
    }
    return 0;
}

static void callback_setup(int type, xen_callback_t *cb)
{
#if 0 == longmode
    cb->cs = fix_sel32(cb->cs);
#endif
    xencb[type] = *cb;
}

static void callback_clear(int type)
{
#if longmode
    xencb[type] = 0;
#else
    xencb[type].cs = 0;
    xencb[type].eip = 0;
#endif
}

static sreg_t set_callbacks(ureg_t *args)
{
#if longmode
    callback_setup(CALLBACKTYPE_event, &args[0]);
    callback_setup(CALLBACKTYPE_failsafe, &args[1]);
#else
    xen_callback_t cb;

    cb.cs  = args[0];
    cb.eip = args[1];
    callback_setup(CALLBACKTYPE_event, &cb);
    cb.cs  = args[2];
    cb.eip = args[3];
    callback_setup(CALLBACKTYPE_failsafe, &cb);
#endif
    return 0;
}

static sreg_t callback_op(ureg_t *args)
{
    struct callback_register cb;

    memcpy_pf(&cb, (void*)(args[1]), sizeof(cb));
    if (cb.type >= 8)
	return -EINVAL;
    
    switch (args[0]) {
    case CALLBACKOP_register:
	callback_setup(cb.type, &cb.address);
	if (cb.flags)
	    printk("%s: FIXME: ignored flags\n", __FUNCTION__);
	break;
    case CALLBACKOP_unregister:
	callback_clear(cb.type);
	break;
    default:
	printk("%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
	return -ENOSYS;
    }
    return 0;
}

static sreg_t set_gdt(ureg_t *args)
{
#if longmode
    uint64_t mfns[16];
#else
    uint32_t mfns[16];
#endif
    uint32_t entries = args[1];
    uint32_t pages = (entries + 511) / 512;
    struct descriptor_32 *src, *dst, tmp;
    int p,e;

    if (0 != memcpy_pf(mfns, (void*)(args[0]), sizeof(mfns)))
	return -EFAULT;
    if (entries > (0xe000 >> 3))
	return -EINVAL;

    for (p = 0; p < pages; p++) {
	gdt_mfn[p] = mfns[p];
	src = map_page(frame_to_addr(mfns[p]));
	dst = xen_gdt + p * 512;
	for (e = 0; e < 512; e++) {
	    tmp = src[e];
	    fix_desc(&tmp);
	    dst[e] = tmp;
	}
	free_page(src);
    }
    return 0;
}
