#include "emu32.h"

/* --------------------------------------------------------------------- */
/* inline asm bits                                                       */

int is_pae(void)
{
    static int pae = -1;

    if (-1 == pae) {
	pae = (read_cr4() & X86_CR4_PAE);
	printk("pae is turned %s\n", pae ? "on" : "off");
    }
    return pae;
}

static inline void vmexit_panic(void)
{
    asm volatile("outb %%al,$0xeb" : : : "memory");
}

/* --------------------------------------------------------------------- */
/* helper functions                                                      */

int panic(char *message)
{
    printk("panic: %s\n", message);
    vmexit_panic();
    return 0;
}

static int bounce_trap(struct regs_32 *regs, int trapno, int cbno)
{
    uint32_t *kesp, eip, cs;
    int stack_switch = 0;
    int error_code = 0;
    int interrupt = 0;
    int k = 0;

    if (trapno >= 0) {
	/* trap bounce */
	eip  = xen_ctxt.trap_ctxt[trapno].address;
	cs   = xen_ctxt.trap_ctxt[trapno].cs;
	if (TI_GET_IF(&xen_ctxt.trap_ctxt[trapno]))
	    interrupt = 1;
	switch (trapno) {
	case 13: /* general protection */
	    error_code = 1;
	    break;
	case 14: /* page fault */
	    shared_info.vcpu_info[0].arch.cr2 = read_cr2();
	    error_code = 1;
	    break;
	}
    }
    if (cbno >= 0) {
	/* callback */
	eip  = xencb[cbno].eip;
	cs   = xencb[cbno].cs;
	switch (cbno) {
	case CALLBACKTYPE_event:
 	    interrupt = 1;
	    break;
	}
    }

    if (!cs)
	panic("no guest trap handler");

    if ((regs->cs & 0x03) < (cs & 0x03))
	panic("bounce trap: illegal ring switch\n");
    if ((regs->cs & 0x03) > (cs & 0x03))
	stack_switch = 1;

    /* prepare guest stack: copy from emu, so the handler
     * jumps straigt back without round-trip via emu */
    if (stack_switch) {
	kesp = (void*)xen_tss.esp1;
	kesp[-(++k)] = regs->ss;            // push ss
	kesp[-(++k)] = regs->esp;           // push esp
    } else {
	kesp = (void*)regs->esp;
    }

    kesp[-(++k)] = regs->eflags;            // push eflags
    kesp[-(++k)] = regs->cs;                // push cs
    kesp[-(++k)] = regs->eip;               // push eip
    if (error_code)
	kesp[-(++k)] = regs->error;         // push error code

    /* prepare emu stack, so iret jumps to the kernels handler. */
    regs->eip     = eip;
    regs->cs      = cs;
    regs->eflags &= EFLAGS_TRAPMASK;
    if (stack_switch) {
	regs->ss  = xen_tss.ss1;
	regs->esp = xen_tss.esp1;
    }
    regs->esp -= 4*k;

    return 0;
}

static void print_gpf_info(struct regs_32 *regs)
{
    uint8_t *code = (void*)regs->eip;
    
    printk("%s: index 0x%x%s%s%s, eflags %x, cs:eip %x:%x "
	   "-> 0x%02x, 0x%02x, 0x%02x, 0x%02x,  0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	   __FUNCTION__, (int)(regs->error >> 3),
	   (regs->error & 0x04) ? ", TI"  : "",
	   (regs->error & 0x02) ? ", IDT" : "",
	   (regs->error & 0x01) ? ", EXT" : "",
	   regs->eflags, regs->cs, regs->eip,
	   code[0], code[1], code[2], code[3],
	   code[4], code[5], code[6], code[7]);
}

static void tss_init(void)
{
    int tr = 0xe040;

    xen_tss.esp0 = (intptr_t)stack_high;
    xen_tss.ss0  = 0xe010;

    xen_gdt[ tr >> 3 ] = mkdesc32((intptr_t)(&xen_tss), sizeof(xen_tss)-1, 0x89, 0);
    asm volatile("ltr %w0" : : "r" (tr));
}

static void idt_init(void)
{
    struct {
	uint16_t len;
	uint32_t ptr;
    } __attribute__((packed)) idtp = {
	.len = sizeof(xen_idt)-1,
	.ptr = (intptr_t)xen_idt,
    };
    int i;

    for (i = 0; i < 256; i++)
	xen_idt[i] = mkgate32(0xe008, (intptr_t)int_unknown, 0x8e);
    
    xen_idt[    0 ] = mkgate32(0xe008, (intptr_t)division_by_zero,    0x8e);
    xen_idt[    3 ] = mkgate32(0xe008, (intptr_t)debug_int3,          0xee);
    xen_idt[    6 ] = mkgate32(0xe008, (intptr_t)illegal_instruction, 0x8e);
    xen_idt[   13 ] = mkgate32(0xe008, (intptr_t)general_protection,  0x8e);
    xen_idt[   14 ] = mkgate32(0xe008, (intptr_t)page_fault,          0x8e);

    xen_idt[ 0x20 ] = mkgate32(0xe008, (intptr_t)event_callback,      0x8e);
    xen_idt[ 0x82 ] = mkgate32(0xe008, (intptr_t)xen_hypercall,       0xae);

    asm volatile("lidt %0" : : "m" (idtp));
}

/* --------------------------------------------------------------------- */

static void print_page_fault_info(struct regs_32 *regs, uint32_t cr2)
{
    printk("%s:%s%s%s%s%s, cs:eip %x:%x, cr2 %x\n", __FUNCTION__,
	   regs->error & 0x01 ? " preset"  : " nopage",
	   regs->error & 0x02 ? " write"   : " read",
	   regs->error & 0x04 ? " user"    : " kernel",
	   regs->error & 0x08 ? " reserved-bit"  : "",
	   regs->error & 0x10 ? " instr-fetch"   : "",
	   regs->cs, regs->eip, read_cr2());
}

static void pf_loopcheck(struct regs_32 *regs, uint32_t cr2)
{
    static uint32_t last_cr2, last_err, loop, noloop;

    if (last_cr2 == cr2 && last_err == regs->error) {
	loop++;
	if (noloop) {
	    printk("%s: start looping (saw %d normal faults)\n", __FUNCTION__, noloop);
	    noloop = 0;
	}
    } else {
	last_cr2 = cr2;
	last_err = regs->error;
	noloop++;
	if (loop) {
	    printk("%s: stopped looping (repeated %d times)\n", __FUNCTION__, loop);
	    printk("--\n");
	    loop = 0;
	}
    }
    if (loop) {
	if (1 == loop)
	    print_page_fault_info(regs, cr2);
	if (loop > 1024)
	    panic("page fault looping");
    }
}

static int pf_fixup_readonly(struct regs_32 *regs, uint32_t cr2)
{
    if (!is_pae()) {
	uint32_t *pte = find_pte_32_lpt(cr2);
	if (cr2 >= XEN_M2P_32)
	    return 0;
	if (*pte & _PAGE_USER)
	    return 0;
	/* is kernel page */
	*pte |= _PAGE_RW;
    } else {
	uint64_t *pte = find_pte_pae_lpt(cr2);
	if (cr2 >= XEN_M2P_PAE)
	    return 0;
	if (*pte & _PAGE_USER)
	    return 0;
#if 1 /* DEBUG -- this should not be called at not-present pages */
	if (!(*pte & _PAGE_PRESENT)) {
	    printk("%s: cr2 %x pte %" PRIx64 "\n", __FUNCTION__, cr2, *pte);
	    panic("linear page table problem here");
	}
#endif
	/* is kernel page */
	*pte |= _PAGE_RW;
    }
    emu_faults[XEN_FAULT_PAGE_FAULT_FIX_RO]++;
    flush_tlb_addr(cr2);
    return 1;
}

static int pf_fixup_extable(struct regs_32 *regs)
{
    uintptr_t *ptr;

    for (ptr = _estart; ptr < _estop; ptr += 2) {
	if (ptr[0] != regs->eip)
	    continue;
#if 0
	printk("fixup: %x -> %x\n", ptr[0], ptr[1]);
#endif
	regs->eip = ptr[1];
	emu_faults[XEN_FAULT_PAGE_FAULT_FIX_EXTAB]++;
	return 1;
    }
    return 0;
}

/* --------------------------------------------------------------------- */
/* called from assembler                                                 */

asmlinkage void do_boot(struct regs_32 regs)
{
    uint32_t cr3_mfn;
    
    printk("this is emu32 (xenner %s)\n", VERSION);
    tss_init();
    idt_init();
    paging_init();
    printk("emu setup done (tss, idt, paging)\n");

    /* guest paging setup */
    cr3_mfn = xen_ctxt.ctrlreg[3] >> PAGE_SHIFT;
    update_emu_mappings(cr3_mfn);
    write_cr3(frame_to_addr(cr3_mfn));
    
    /* guest register setup */
    regs.eax    = xen_ctxt.user_regs.eax;
    regs.ebx    = xen_ctxt.user_regs.ebx;
    regs.ecx    = xen_ctxt.user_regs.ecx;
    regs.edx    = xen_ctxt.user_regs.edx;
    regs.esi    = xen_ctxt.user_regs.esi;
    regs.edi    = xen_ctxt.user_regs.edi;
    regs.ebp    = xen_ctxt.user_regs.ebp;
    regs.eip    = xen_ctxt.user_regs.eip;
    regs.cs     = xen_ctxt.user_regs.cs;
    regs.eflags = xen_ctxt.user_regs.eflags;
    regs.esp    = xen_ctxt.user_regs.esp;
    regs.ss     = xen_ctxt.user_regs.ss;

    asm volatile("mov %0, %%ds;\n" :: "r" (xen_ctxt.user_regs.ds) : "memory");
    asm volatile("mov %0, %%es;\n" :: "r" (xen_ctxt.user_regs.es) : "memory");
    asm volatile("mov %0, %%fs;\n" :: "r" (xen_ctxt.user_regs.fs) : "memory");
    asm volatile("mov %0, %%gs;\n" :: "r" (xen_ctxt.user_regs.gs) : "memory");

    /* jump to kernel (via iret in asm code calling us) */
    printk("booting kernel (entry %x:%x) ...\n", regs.cs, regs.eip);
}

asmlinkage void do_illegal_instruction(struct regs_32 regs)
{
    int skip;

    emu_faults[XEN_FAULT_ILLEGAL_INSTRUCTION]++;
    if (context_is_emu(&regs))
	panic("ring0 (emu) illegal instruction");
    if (context_is_user(&regs)) {
	bounce_trap(&regs, 6, -1);
	return;
    }
    
    skip = emulate_32(&regs);
    switch (skip) {
    case -1: /* error */
	panic("instruction emulation failed\n");
	break;
    case 0:  /* bounce to guest */
	bounce_trap(&regs, 6, -1);
	break;
    default: /* handled */
	regs.eip += skip;
	break;
    }
}

asmlinkage void do_general_protection(struct regs_32 regs)
{
    int skip;

    emu_faults[XEN_FAULT_GENERAL_PROTECTION]++;
    if (context_is_emu(&regs)) {
	print_gpf_info(&regs);
	panic("ring0 (emu) general protection fault");
    }
    if (context_is_user(&regs)) {
	print_gpf_info(&regs);
	emu_faults[XEN_FAULT_GENERAL_PROTECTION_GUEST]++;
	bounce_trap(&regs, 13, -1);
	return;
    }
    
    if (0 != regs.error) {
	print_gpf_info(&regs);
	panic("unhandled kernel gpf");
    }

    skip = emulate_32(&regs);
    switch (skip) {
    case -1: /* error */
	panic("instruction emulation failed\n");
	break;
    case 0:  /* bounce to guest */
	emu_faults[XEN_FAULT_GENERAL_PROTECTION_GUEST]++;
	bounce_trap(&regs, 13, -1);
	break;
    default: /* handled */
	emu_faults[XEN_FAULT_GENERAL_PROTECTION_EMUINS]++;
	regs.eip += skip;
	break;
    }
}

asmlinkage void do_page_fault(struct regs_32 regs)
{
    uint32_t cr2 = read_cr2();

    if (0)
	pf_loopcheck(&regs, cr2);

    emu_faults[XEN_FAULT_PAGE_FAULT]++;
    if (context_is_emu(&regs)) {
#if 0
	/* fails for linear page table faults */
	if (regs.error == 3) {
	    print_page_fault_info(&regs);
	    if (pf_fixup_readonly(&regs, cr2))
		return;
	}
#endif
	if (pf_fixup_extable(&regs))
	    return;
	print_page_fault_info(&regs, cr2);
	panic("ring0 (emu) page fault");
    }

    if (regs.error == 3) {
	/* kernel write to r/o page */
	if (pf_fixup_readonly(&regs, cr2))
	    return;
    }
    
    emu_faults[XEN_FAULT_PAGE_FAULT_GUEST]++;
    bounce_trap(&regs, 14, -1);
}

asmlinkage void do_event_callback(struct regs_32 regs)
{
    if (shared_info.vcpu_info[0].evtchn_upcall_mask) {
	/* FIXME */
	printk("%s: IRQs off\n", __FUNCTION__);
	return;
    }
    emu_faults[XEN_FAULT_EVENT_CALLBACK]++;
    shared_info.vcpu_info[0].evtchn_upcall_mask = 1;
    bounce_trap(&regs, -1, CALLBACKTYPE_event);
}

asmlinkage void do_guest_forward(struct regs_32 regs)
{
    bounce_trap(&regs, regs.trapno, -1);
}

asmlinkage void do_debug_int3(struct regs_32 regs)
{
    printk("%s: cs:eip %04x:%08x -- eax %08x esi %08x edi %08x ebp %08x\n",
	   __FUNCTION__, regs.cs, regs.eip,
	   regs.eax, regs.esi, regs.edi, regs.ebp);
}
